/*************************************************************************/
/*                                                                       */
/* 8hz-mp3: Multiplatform MPEG 1, Layer 3 encoder.                       */
/*                                                                       */
/* May 1998: 8hz, added WAV support, optmized calculations, added some   */
/*           precalc tables. Gained 60-70% over ISO implementation.      */
/*                                                                       */
/*                                                                       */
/*                                                                       */
/*                                                                       */
/*                                                                       */
/*                                                                       */
/*                                                                       */
/*************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#ifndef _WIN32
#include <unistd.h>
#endif	/*	_WIN32	*/
#include <time.h>

#include "types.h"
#include "error.h"
#include "wave.h"
#include "layer3.h"

config_t config;

#ifdef MFCPCM
int pcm_switch = 0;
int pcm_rate = 0;
int pcm_byte_order = 0; /* assume LOHI */
#endif /* MFCPCM */


static void print_header()
{
    fprintf(stderr,"8hz-mp3 1998 8Hz software http://www.8hz.com\n");
#ifdef _WIN32	/*	support 'L3Enc.exe' compatible parameter for Win32(Win32 original).	by Jun	*/
    fprintf(stderr,"for Win32(03J) 1998 Jun Somekawa http://www.yk.rim.or.jp/~somekawa/\n");
#endif	/*	_WIN32	*/
}

static void print_usage()
{
    fprintf(stderr,"USAGE   :  8hz-mp3 [options] <infile> <outfile>\n");
    fprintf(stderr,"OPTIONS : -h            this help message\n");
    fprintf(stderr,"          -b <bitrate>  set the bitrate, default 128kbit\n");
    fprintf(stderr,"          -c            set copyright flag, default off\n");
    fprintf(stderr,"          -o            set original flag, default off\n");
#ifdef MFCPCM
    fprintf(stderr,"          -r            force raw PCM reading instead of wav\n");
    fprintf(stderr,"          -f <Hz>       set frequency for PCM reading. Default: 44100\n");
    fprintf(stderr,"          -x            pcm is bigEndian byte order (default: littleEndian)\n");
/*  fprintf(stderr,"          -q <int>      Quality switch. (Default : 0. highest quality) [Normal range is 0.. 30]\n");	*/
    fprintf(stderr,"          -q <int>      Quality switch. (Default : 0. highest quality)\n");
    fprintf(stderr,"                        [Normal range is 0.. 30]\n");
#endif /* MFCPCM */
#ifdef _WIN32	/*	support 'L3Enc.exe' compatible parameter for Win32(Win32 original).	by Jun	*/
    fprintf(stderr,"          ==== L3Enc.exe compatible parameters for Win32 ====\n");
    fprintf(stderr,"          -br <bitrate> set the bitrate, default 112kbit\n");
    fprintf(stderr,"          -dm           set downmix for mono, default off(cannot use)\n");
    fprintf(stderr,"          -crc          set CRC flag, default off(cannot use)\n");
    fprintf(stderr,"          -hq           set High quality output, default off(cannot use)\n");
#endif	/*	_WIN32	*/
    fprintf(stderr,"\n"); 
    fprintf(stderr,"Use - as infile to read from stdin\n");
    fprintf(stderr,"\n"); 
}

static void set_defaults()
{
    config.byte_order = DetermineByteOrder();
    if(config.byte_order==order_unknown) ERROR("Can't determine byte order");

    config.mpeg.type = TYPE_MPEG_I;
    config.mpeg.layr = LAYR_III;
    config.mpeg.mode = MODE_STEREO;
    config.mpeg.bitr = 128;
    config.mpeg.psyc = PSYC_ATT;
    config.mpeg.emph = EMPH_NONE; 
    config.mpeg.crc  = 0;
    config.mpeg.ext  = 0;
    config.mpeg.mode_ext  = 0;
    config.mpeg.copyright = 0;
    config.mpeg.original  = 0;  
    config.qual = 0;  /* 0 is highest quality. */
}

static bool parse_command(int argc, char** argv)
{
    int i = 0;
#ifndef	_WIN32
	int	ch;
#endif

    if(argc<3) return false;

#ifdef _WIN32	/*	support 'L3Enc.exe' compatible parameter for Win32(Win32 original).	by Jun	*/
	config.infile = config.outfile = NULL;	/*	Filename initialized	*/
	config.from_stdin = 0;
	while (++i < argc)
	{
		if (argv[i][0] == '-')
		{
			if (strncmp(&argv[i][1], "br", 2) == 0)
			{
				/*	"-br"	Setting bitrate	*/
				if (argv[i + 1][0] != '-')
				{
					config.mpeg.bitr = atoi(argv[++i]) / 1000;
					if (config.mpeg.bitr == 0)
						config.mpeg.bitr = 112;	/*	L3Enc.exe's default 112kbit/sec.	*/
				}
				else
					config.mpeg.bitr = 112;		/*	L3Enc.exe's default 112kbit/sec.	*/
				config.mpeg.original  = 1;		/*	If L3Enc.exe mode, set original flag. 	*/
			}
			else
			if (strncmp(&argv[i][1], "dm", 2) == 0)
			{
				/*	"-dm"	Downmix to mono mode	*/
				/*	8hz-mp3 cannot use this mode now.	*/
			}
			else
			if (strncmp(&argv[i][1], "crc", 3) == 0)
			{
				/*	"-crc"	CRC	*/
				/*	config.mpeg.crc = 1;	*/	/*	CRC on buged?	*/
			}
			else
			if (strncmp(&argv[i][1], "hq", 2) == 0)
			{
				/*	"-hq"	High Quality mode		*/
				/*	8hz-mp3 cannot use this mode.	*/
			}
			else
			{
				/*	8hz-mp3 original parameters	*/
				switch(argv[i][1])
				{
					case '\0': config.from_stdin = 1;
							   break;		/*	stdin cannot use now	*/
					case 'b' : config.mpeg.bitr = atoi(argv[++i]);
							   if (config.mpeg.bitr == 0)
								   config.mpeg.bitr = 128;	/*	8hz-mp3's default 128kbit/sec.	*/
							   break;
					case 'c' : config.mpeg.copyright = 1;
							   break;
					case 'o' : config.mpeg.original  = 1;
							   break;
#ifdef MFCPCM
					case 'r' : pcm_switch = 1;
							   break;
					case 'f' : pcm_rate = /* atoi(optarg) */atoi(argv[++i]);
							   break;
					case 'x' : pcm_byte_order = 1;
						   break;
					case 'q' : config.qual=/*atoi(optarg) */atoi(argv[++i]);
							   if ((config.qual<0) || (config.qual>100)) {
						 printf("Quality error: %i\n",config.qual);
						 exit(1);
						   }
						   break;
#endif /* MFCPCM */
					case 'h' :
					default  : return false;
			   }
			}
		}
		else if (config.from_stdin == 0 && config.infile == NULL)
			config.infile = argv[i];	/*	First is input-file(*.wav)	*/
		else
			config.outfile = argv[i];	/*	Next is output-file(*.mp3)	*/
	}
#else
    while ((ch = getopt(argc, argv, "b:corf:xq:")) != EOF) 
        switch((char)ch)
        {
            case 'b' : config.mpeg.bitr = atoi(optarg);
                       break;
            case 'c' : config.mpeg.copyright = 1;
                       break;
            case 'o' : config.mpeg.original  = 1;
                       break;
#ifdef MFCPCM
	    case 'r' : pcm_switch = 1;
	               break;
	    case 'f' : pcm_rate = atoi(optarg);
	               break;
	    case 'x' : pcm_byte_order = 1;
		       break;
	    case 'q' : config.qual=atoi(optarg);
	               if ((config.qual<0) || (config.qual>100)) {
			 printf("Quality error: %i\n",config.qual);
			 exit(1);
			   }
		       break;
#endif /* MFCPCM */
            case 'h' :
            default  : return false;
       }

    argc -= optind;
    argv += optind;
    if (argc!=2)
      return false;

    if (!strcmp(argv[0],"-")) {
      config.from_stdin=1;
    } else {
      config.from_stdin=0;
      config.infile = argv[0];
    }

    config.outfile = argv[1];
#endif	/*	_WIN32	*/
    return true;
}

static int find_samplerate_index(long freq)
{
    static long mpeg1[3] = {44100, 48000, 32000};
    static long mpeg2[3] = {22050, 24000, 16000};
    long *table;
    int i;

    table = (config.mpeg.type==TYPE_MPEG_I)?mpeg1:mpeg2;
    for(i=0;i<3;i++)
        if(freq==table[i]) return i;

    ERROR("Invalid samplerate");
    return -1;
}

static int find_bitrate_index(int bitr)
{
    static long mpeg1[15] = {0,32,40,48,56,64,80,96,112,128,160,192,224,256,320};
    static long mpeg2[15] = {0, 8,16,24,32,40,48,56, 64, 80, 96,112,128,144,160};
    long *table;
    int i;

    table = (config.mpeg.type==TYPE_MPEG_I)?mpeg1:mpeg2;
    for(i=0;i<15;i++)
        if(bitr==table[i]) return i;

    ERROR("Invalid bitrate");
    return -1;
}

static void check_config()
{
    static char *mode_names[4]    = { "stereo", "j-stereo", "dual-ch", "mono" };
    static char *layer_names[3]   = { "I", "II", "III" };
    static char *version_names[2] = { "MPEG-II (LSF)", "MPEG-I" };
    static char *psy_names[3]     = { "", "MUSICAM", "AT&T" };
    static char *demp_names[4]    = { "none", "50/15us", "", "CITT" };

    config.mpeg.samplerate_index = find_samplerate_index(config.wave.samplerate);
    config.mpeg.bitrate_index    = find_bitrate_index(config.mpeg.bitr);

   printf("%s layer %s, %s  Psychoacoustic Model: %s\n",
           version_names[config.mpeg.type],
           layer_names[config.mpeg.layr], 
           mode_names[config.mpeg.mode],
           psy_names[config.mpeg.psyc]);
   printf("Bitrate=%d kbps  ",config.mpeg.bitr );
   printf("De-emphasis: %s  CRC: %s %s %s\n",
          demp_names[config.mpeg.emph], 
          ((config.mpeg.crc)?"on":"off"),
          ((config.mpeg.original)?"Original":""),
          ((config.mpeg.copyright)?"(C)":""));

}

int main(int argc, char **argv)
{
    time_t end_time;

    time(&config.start_time);
    print_header();
    set_defaults();
    if(!parse_command(argc,argv)) { print_usage(); return -1; }
    if(!wave_open()) ERROR("Unable to open input file...");
	if (config.wave.channels == 1)
		config.mpeg.mode = MODE_MONO;	/*	If WAVE file is Monaural, set Monauralmode. */
    check_config();

    if ( config.from_stdin )
        printf("Encoding stdin to \"%s\"\n", config.outfile);
    else
        printf("Encoding \"%s\" to \"%s\"\n", config.infile, config.outfile);
    switch(config.mpeg.layr)
    {
        case LAYR_I  :
        case LAYR_II :
        default      :
            ERROR("Layer not supported");
        case LAYR_III:
            L3_compress();
            break;
    }
    wave_close();

    time(&end_time);
    end_time -= config.start_time;
    fprintf(stdout," Finished in %2ld:%2ld:%2ld\n",
            end_time/3600,(end_time/60)%60,end_time%60);
    return 0;
} 
